/*****************************************************************************
* Copyright (C) 2000 Nik Weidenbacher
* nikw@martnet.com
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************************************************************/

#include "config.h"
#include "ehnt.h"

/* ProcessFlow is my least favorite part of ehnt.  It needs to be made pretty 
   and sensible. 

   Note for developers:  when ProcessFlow is called, the flow record in 
   flow_ver5_rec has some fields in network byte order and others in host
   byte order.  They should maybe be all host order some day to save sanity.
*/
int NewStats(struct ehnt_stats * );

int ProcessFlow(struct flow_ver5_rec * rec, struct in_addr * routeraddr,
		struct ehnt_struct * my_ehnt_struct) {

  char * routeraddr_string;
  static double last_packetcount;
  static double local_flowcount=0;
  struct in_addr srcip,dstip,nexthop;
  struct ehnt_config * e_cfg;
  struct ehnt_filter * e_flt;
  struct ehnt_data * e_dat;
  struct ehnt_stats * stats, * old_stats;

  e_cfg=my_ehnt_struct->cfg;
  e_flt=my_ehnt_struct->flt;
  e_dat=my_ehnt_struct->dat;
  stats=e_dat->stats;
  old_stats=e_dat->old_stats;
  
  routeraddr_string=inet_ntoa(*routeraddr);  /* remember, inet_ntoa wants network byte order */

  if (e_cfg->mode == EM_TOP) { /* if mode is top, we may need to generate a report. */
    time_t mr_t;
    struct tm * ts;
    int currentmin;

    /* first, (only if this flow is from a new packet) see if this a new time interval.
       if it is, display stats from the last 
       interval, and clear out AS counters.  Since we don't know how long we were blocking
       for.  But I'm kind of assuming that we were blocking for less than 1 interval. */
    if (e_dat->packetcount != last_packetcount) {
      mr_t=time(NULL);
      ts=localtime(&mr_t);    
      currentmin=(ts->tm_min + (ts->tm_hour*60) + (ts->tm_yday * 60 * 24));

      /* if this is a new interval, display the report for the previous interval */
      if (  ((currentmin%e_cfg->si) == 0)  && ( currentmin != e_dat->recentmin)) {	
	time_t interval;
	interval=mr_t-e_dat->last_t;  /* amount of time passed in seconds since stats were last cleared */
	e_dat->last_t=mr_t;           /* last_t is used in addition to recentmin for seconds resolution */
	e_dat->recentmin=currentmin;

	if (! old_stats) {  /* create the old_stats struct if necessary */
	  e_dat->old_stats=malloc(sizeof(struct ehnt_stats));
	  old_stats=e_dat->old_stats;
	} else { /* if the struct exists, clean it out */
	  FreeStats(old_stats);
	}
	old_stats->interval=interval;
	old_stats->time=mr_t;
	/* copy current stats to old stats */
	old_stats->InBoundTotal=stats->InBoundTotal;
	old_stats->OutBoundTotal=stats->OutBoundTotal;

	old_stats->InBoundSrcAS=stats->InBoundSrcAS;
	old_stats->InBoundDstAS=stats->InBoundDstAS;
	old_stats->OutBoundSrcAS=stats->OutBoundSrcAS;
	old_stats->OutBoundDstAS=stats->OutBoundDstAS;

	old_stats->InBoundProto=stats->InBoundProto;
	old_stats->OutBoundProto=stats->OutBoundProto;

	old_stats->InBoundSrcTcpPort=stats->InBoundSrcTcpPort;
	old_stats->InBoundDstTcpPort=stats->InBoundDstTcpPort;
	old_stats->OutBoundSrcTcpPort=stats->OutBoundSrcTcpPort;
	old_stats->OutBoundDstTcpPort=stats->OutBoundDstTcpPort;
	
	old_stats->InBoundSrcUdpPort=stats->InBoundSrcUdpPort;
	old_stats->InBoundDstUdpPort=stats->InBoundDstUdpPort;
	old_stats->OutBoundSrcUdpPort=stats->OutBoundSrcUdpPort;
	old_stats->OutBoundDstUdpPort=stats->OutBoundDstUdpPort;
	
	/* reset the current stats (mallocs new stats tables) */
	NewStats(stats);

	ShowReport(old_stats,my_ehnt_struct);
      }
      last_packetcount=e_dat->packetcount;
    }
  } /* EM_TOP */

  /* This big nasty if condition is the flow filter. */
  if (   ( (! e_flt->AS) || (e_flt->AS==rec->src_as) || (e_flt->AS==rec->dst_as) ) &&
	 ( (! e_flt->port) || (e_flt->port==rec->srcport) || (e_flt->port==rec->dstport) ) &&
	 ( (! e_flt->prefix) || (e_flt->prefix->s_addr==(rec->srcaddr & e_flt->pfxmask->s_addr)) ||
	   (e_flt->prefix->s_addr==(rec->dstaddr & e_flt->pfxmask->s_addr) ) ) && 
	 ( (! e_flt->router) || (e_flt->router->s_addr==routeraddr->s_addr) ) &&
	 ( (! e_flt->proto) || (e_flt->proto==rec->prot)) &&
	 ( (! e_flt->big) || ( rec->dOctets > e_dat->maxbytes) || ( rec->dPkts > e_dat->maxpackets) )
	 )
    {
      /* If the flow passes the filter, act on it according to the current display mode */
      local_flowcount++;
      if (e_cfg->mode == EM_TOP) {
	/* we can tell inbound from outbound if the user has specified an ifIndex */
	if (e_flt->intnum) { 
	  if (rec->output_index == e_flt->intnum) {
	    stats->OutBoundTotal+=rec->dOctets;
	    stats->OutBoundSrcAS[rec->src_as]+=rec->dOctets;
	    stats->OutBoundDstAS[rec->dst_as]+=rec->dOctets;
	    stats->OutBoundProto[rec->prot]+=rec->dOctets;
	    if (rec->prot == 6) { /*TCP*/
	      stats->OutBoundSrcTcpPort[rec->srcport]+=rec->dOctets;
	      stats->OutBoundDstTcpPort[rec->dstport]+=rec->dOctets;
	    } else if (rec->prot == 17) { /*UDP*/
	      stats->OutBoundSrcUdpPort[rec->srcport]+=rec->dOctets;
	      stats->OutBoundDstUdpPort[rec->dstport]+=rec->dOctets;
	    }
	  }
	}
	/* inbound on an interface, or the user hasn't specified an ifIndex and
	   we just put everthing in the InBound tables:  */
	if ( (! e_flt->intnum) || (rec->input_index == e_flt->intnum) ) {
	  stats->InBoundTotal+=rec->dOctets;
	  stats->InBoundSrcAS[rec->src_as]+=rec->dOctets;
	  stats->InBoundDstAS[rec->dst_as]+=rec->dOctets;
	  stats->InBoundProto[rec->prot]+=rec->dOctets;
	  if (rec->prot == 6) { /*TCP*/
	    stats->InBoundSrcTcpPort[rec->srcport]+=rec->dOctets;
	    stats->InBoundDstTcpPort[rec->dstport]+=rec->dOctets;
	  } else if (rec->prot == 17) { /*UDP*/
	    stats->InBoundSrcUdpPort[rec->srcport]+=rec->dOctets;
	    stats->InBoundDstUdpPort[rec->dstport]+=rec->dOctets;
	  }
	}
	if (rec->dOctets > e_dat->maxbytes) 
	  {e_dat->maxbytes=rec->dOctets; e_dat->maxbytesAS1=rec->src_as; e_dat->maxbytesAS2=rec->dst_as; }
	if (rec->dPkts> e_dat->maxpackets)  
	  {e_dat->maxpackets=rec->dPkts; e_dat->maxpacketsAS1=rec->src_as; e_dat->maxpacketsAS2=rec->dst_as; }
      } /* EM_TOP */
      else if (e_cfg->mode == EM_SHORTDUMP) {

	srcip.s_addr=rec->srcaddr;
	dstip.s_addr=rec->dstaddr;
	nexthop.s_addr=rec->nexthop;

	printf("%.0f: %-15s ",local_flowcount,routeraddr_string);
	printf("%2d/%-15.15s:%-5d ",
	       rec->input_index,
	       inet_ntoa(srcip),
	       rec->srcport);
	printf("-> %2d/%-15.15s:%-5d ",
	       rec->output_index,
	       inet_ntoa(dstip),
	       rec->dstport);
	
	printf("\n b/p:%s/%-6d proto %2d",prettybytes(rec->dOctets),rec->dPkts,rec->prot);
	printf(" %17s(%5d)->",ASN_Lookup(rec->src_as),rec->src_as);
	printf("%17s(%5d)\n",ASN_Lookup(rec->dst_as),rec->dst_as);

	if (e_flt->big) {
	  if (rec->dOctets > e_dat->maxbytes) 
	    {e_dat->maxbytes=rec->dOctets; e_dat->maxbytesAS1=rec->src_as; e_dat->maxbytesAS2=rec->dst_as; }
	  if (rec->dPkts> e_dat->maxpackets)
	    {e_dat->maxpackets=rec->dPkts; e_dat->maxpacketsAS1=rec->src_as; e_dat->maxpacketsAS2=rec->dst_as; }
	}

      } /* EM_SHORTDUMP */
      else if (e_cfg->mode == EM_DUMP) {
	srcip.s_addr=rec->srcaddr;
	dstip.s_addr=rec->dstaddr;
	nexthop.s_addr=rec->nexthop;

	printf("flow #%.0f received from router %s, IP protocol %d\n",
	       local_flowcount,routeraddr_string,rec->prot);
	printf("  input ifIndex:     %d\n",rec->input_index);
	printf("  source IP address: %s\n",inet_ntoa(srcip));
	printf("  source port:       %d\n",rec->srcport);
	printf("  source AS:         %s(%d)\n",ASN_Lookup(rec->src_as),rec->src_as);
	printf("  output ifIndex:    %d\n",rec->output_index);
	printf("  dest IP address:   %s\n",inet_ntoa(dstip));
	printf("  dest port:         %d\n",rec->dstport);
	printf("  dest AS:           %s(%d)\n",ASN_Lookup(rec->dst_as),rec->dst_as);
	printf("  bytes in flow:     %s\n", prettybytes(rec->dOctets));
	printf("  packets in flow:   %d\n", rec->dPkts);

	if (e_flt->big) {
	  if (rec->dOctets > e_dat->maxbytes) 
	    {e_dat->maxbytes=rec->dOctets; e_dat->maxbytesAS1=rec->src_as; e_dat->maxbytesAS2=rec->dst_as; }
	  if (rec->dPkts> e_dat->maxpackets)
	    {e_dat->maxpackets=rec->dPkts; e_dat->maxpacketsAS1=rec->src_as; e_dat->maxpacketsAS2=rec->dst_as; }
	}

      } /* EM_DUMP */
      else if (e_cfg->mode == EM_COLONDUMP) {

	srcip.s_addr=rec->srcaddr;
	dstip.s_addr=rec->dstaddr;
	nexthop.s_addr=rec->nexthop;

	printf("%.0f:%s",local_flowcount,routeraddr_string);
	printf("%d:%s:%d:",
	       rec->input_index,
	       inet_ntoa(srcip),
	       rec->srcport);
	printf("%d:%s:%d:",
	       rec->output_index,
	       inet_ntoa(dstip),
	       rec->dstport);
	
	printf("%d:%d:%d:",rec->dOctets,rec->dPkts,rec->prot);
	printf("%d:%d\n",rec->src_as,rec->dst_as);

	if (e_flt->big) {
	  if (rec->dOctets > e_dat->maxbytes) 
	    {e_dat->maxbytes=rec->dOctets; e_dat->maxbytesAS1=rec->src_as; e_dat->maxbytesAS2=rec->dst_as; }
	  if (rec->dPkts> e_dat->maxpackets)
	    {e_dat->maxpackets=rec->dPkts; e_dat->maxpacketsAS1=rec->src_as; e_dat->maxpacketsAS2=rec->dst_as; }
	}

      } /* EM_COLONDUMP */
  } /* if <packet match> */

  fflush(stdout);
  /* If a flow limit is set, check for it. */
  if ((e_cfg->flowlimit) && (local_flowcount >= e_cfg->flowlimit)) {
    printf("Flow limit reached (%d).  Exiting\n",e_cfg->flowlimit);
    set_term(SET_TERM_COOKED);
    exit(0);
  }
}

int NewStats(struct ehnt_stats * stats) {
  stats->InBoundSrcAS=malloc(AS_TS);
  stats->InBoundDstAS=malloc(AS_TS);
  stats->OutBoundSrcAS=malloc(AS_TS);
  stats->OutBoundDstAS=malloc(AS_TS);

  stats->InBoundProto=malloc(PROTO_TS);
  stats->OutBoundProto=malloc(PROTO_TS);

  stats->InBoundSrcTcpPort=malloc(TCPPORT_TS);
  stats->InBoundDstTcpPort=malloc(TCPPORT_TS);
  stats->OutBoundSrcTcpPort=malloc(TCPPORT_TS);
  stats->OutBoundDstTcpPort=malloc(TCPPORT_TS);

  stats->InBoundSrcUdpPort=malloc(UDPPORT_TS);
  stats->InBoundDstUdpPort=malloc(UDPPORT_TS);
  stats->OutBoundSrcUdpPort=malloc(UDPPORT_TS);
  stats->OutBoundDstUdpPort=malloc(UDPPORT_TS);

  memset(stats->InBoundSrcAS,0,AS_TS);
  memset(stats->InBoundDstAS,0,AS_TS);
  memset(stats->OutBoundSrcAS,0,AS_TS);
  memset(stats->OutBoundDstAS,0,AS_TS);

  memset(stats->InBoundProto,0,PROTO_TS);
  memset(stats->OutBoundProto,0,PROTO_TS);

  memset(stats->InBoundSrcTcpPort,0,TCPPORT_TS);
  memset(stats->InBoundDstTcpPort,0,TCPPORT_TS);
  memset(stats->OutBoundSrcTcpPort,0,TCPPORT_TS);
  memset(stats->OutBoundDstTcpPort,0,TCPPORT_TS);

  memset(stats->InBoundSrcUdpPort,0,UDPPORT_TS);
  memset(stats->InBoundDstUdpPort,0,UDPPORT_TS);
  memset(stats->OutBoundSrcUdpPort,0,UDPPORT_TS);
  memset(stats->OutBoundDstUdpPort,0,UDPPORT_TS);

  stats->InBoundTotal=0;
  stats->OutBoundTotal=0;
}

int FreeStats (struct ehnt_stats *stats) {

  free(stats->InBoundSrcAS);
  free(stats->InBoundDstAS);
  free(stats->OutBoundSrcAS);
  free(stats->OutBoundDstAS);
  
  free(stats->InBoundProto);
  free(stats->OutBoundProto);

  free(stats->InBoundSrcTcpPort);
  free(stats->InBoundDstTcpPort);
  free(stats->OutBoundSrcTcpPort);
  free(stats->OutBoundDstTcpPort);

  free(stats->InBoundSrcUdpPort);
  free(stats->InBoundDstUdpPort);
  free(stats->OutBoundSrcUdpPort);
  free(stats->OutBoundDstUdpPort);
}
