/*****************************************************************************
* Copyright (C) 2000 Nik Weidenbacher
* nikw@martnet.com
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************************************************************/

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/time.h>  
#include <netinet/in.h>
#include <arpa/inet.h> 
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <netinet/tcp.h>
#include <syslog.h>
#include "config.h"

#define MAXCLIENTS 256
#define MAXPACKET 8192

int SetUDPOpts(int);
int SetTCPOpts(int);

int debug;
int use_syslog;

void log_error(char *);

int main(int argc, char **argv) {
 char ips[100];
 int c,i;

 int nf_fd;  /*NetFlow fd*/
 int l_fd;   /*listen fd, for client connections*/
 struct sockaddr_in nf_servaddr, nf_cliaddr, l_servaddr, l_cliaddr;
 socklen_t nf_clilen,l_clilen;
 char nf_packet[MAXPACKET];
 int l_clients[MAXCLIENTS];
 int packetcounter=0;
 extern char *optarg;
 int c_udpport, c_tcpport;

 debug=0;
 use_syslog=0;

 c_udpport=DEFAULT_UDP_PORT;
 c_tcpport=DEFAULT_TCP_PORT;

 while( (c=getopt(argc,argv,"u:t:d")) != -1) {

   switch(c) {
   case 'u':
     sscanf(optarg,"%d",&c_udpport);
     break;
   case 't':
     sscanf(optarg,"%d",&c_tcpport);
     break;
   case 'd':
     debug=1;
     break;
   default:
     printf("Usage: %s [ -u <udpport> ] [ -t <tcpport> ]\n",argv[0]);
     printf("  -u <udpport>  Listen on UDP port <udpport> for netflow packets.  Default %d\n",DEFAULT_UDP_PORT);
     printf("  -t <tcpport>  Listen on TCP port <tcpport> for client connections.  Default %d\n",DEFAULT_TCP_PORT);
     printf("  -d            Debug.  Don't become a daemon, displays all messages on stdout,");
     printf("                displays lots of extra information\n");
     exit(0);
   }
   
 }
 

 /* set up the netflow udp socket */
 if ( (nf_fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
   { log_error("socket (udp) error"); exit(0); }
 memset(&nf_servaddr,0,sizeof(nf_servaddr));
 nf_servaddr.sin_family      = AF_INET;
 nf_servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
 nf_servaddr.sin_port        = htons(c_udpport);
 if ( bind(nf_fd, (struct sockaddr *) &nf_servaddr, sizeof(nf_servaddr)) < 0)
   { log_error("bind UDP error"); exit(0); }

 SetUDPOpts(nf_fd);

 nf_clilen = (socklen_t) sizeof(nf_cliaddr);
 l_clilen = (socklen_t) sizeof(l_cliaddr);
 
 /* set up the listening tcp socket*/
 if ( (l_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   { log_error("socket (tcp) error"); exit(0); }
 memset(&l_servaddr,0,sizeof(l_servaddr));
 l_servaddr.sin_family      = AF_INET;
 l_servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
 l_servaddr.sin_port        = htons(c_tcpport);
 {
  int one=1;
  setsockopt(l_fd, SOL_SOCKET, SO_REUSEADDR, (void*)&one, sizeof(one));
 }
 if ( bind(l_fd,(struct sockaddr *) &l_servaddr, sizeof(l_servaddr)) < 0)
   { log_error ("bind TCP error"); exit(0); }
 if ( listen(l_fd, 32) < 0)
   { log_error("listen error"); exit(0); }
 printf("Listening on UDP port %d and TCP port %d (with a backlog of 32)\n",c_udpport,c_tcpport);
 memset(l_clients,0,sizeof(l_clients));

 /* OK, all the network stuff worked (I think).  Now become a daemon (unless we got -d)*/
 if (! debug) {
   pid_t pid;

   printf("Forking into background.  Future messages will be sent to syslog\n");
   use_syslog=1;

   if ( (pid=fork()) < 0) {
     log_error("fork error"); exit(1);
   } else if (pid != 0) exit(0);

   setsid();
   chdir("/");
   umask(0);

   signal(SIGPIPE,SIG_IGN);
   signal(SIGWINCH,SIG_IGN);
 }
 
 for ( ; ;) {
  fd_set read_fds, write_fds;
  int maxfd;
  int nf_rcount;

  FD_ZERO(&read_fds);

  FD_SET(nf_fd,&read_fds);
  FD_SET(l_fd,&read_fds);
  maxfd=nf_fd;
  if (l_fd >maxfd) { maxfd=l_fd; }

  for (i=0; i<MAXCLIENTS; i++) {
    if (l_clients[i] > 0) {
      FD_SET(l_clients[i],&read_fds);
      if (l_clients[i] > maxfd) { maxfd=l_clients[i]; }
    }
  }

  if ( (c=select(maxfd+1,&read_fds,NULL,NULL,NULL)) < 0)
    { log_error("select (read) error"); exit(0); }
  
  if (FD_ISSET(nf_fd,&read_fds)) {  /*a netflow message is waiting*/

    struct timeval timeout;


    memset(nf_packet,0,sizeof(nf_packet));
    /* new packet */
    if ( (nf_rcount=recvfrom(nf_fd,(char *)(nf_packet + sizeof(nf_cliaddr.sin_addr.s_addr)),
			     MAXPACKET,0,(struct sockaddr *)&nf_cliaddr,&nf_clilen)) < 0)
      { log_error("recvfrom error"); exit(0); }
    
    /* Here we use a non-blocking select to find out which tcp clients we 
       can write to.  This ensures the write won't block, which would of course
       Be Very Bad.  (there is a limit to how much we can write.  it's
       probably 2048 bytes on most unixes.  on linux it's 1 byte, and can't be changed
       see calls to setsockopt for SO_SNDLOWAT below.
    */

    maxfd=0;
    FD_ZERO(&write_fds);
    for (i=0; i<MAXCLIENTS; i++) {
      if (l_clients[i] > 0) {
	FD_SET(l_clients[i],&write_fds);
	if (l_clients[i] > maxfd) { maxfd=l_clients[i]; }
      }
    }

    timeout.tv_sec=0;
    timeout.tv_usec=0;
    if ( (c=select(maxfd+1,NULL,&write_fds,NULL,&timeout)) < 0)
      { log_error("select (write) error"); exit(0); }
    for (i=0; i<MAXCLIENTS; i++) {
      if (l_clients[i] > 0) {
	if (FD_ISSET(l_clients[i],&write_fds)) {
	  memcpy(nf_packet,(void*)&nf_cliaddr.sin_addr.s_addr,sizeof(nf_cliaddr.sin_addr.s_addr));
	  c=write(l_clients[i],&nf_packet,nf_rcount+sizeof(nf_cliaddr.sin_addr.s_addr));
	  if (c < 1) { 
	    if (debug) {
	      printf("client %d/%d seems to have gone away.\n",l_clients[i],i);
	    } else {
	      char msg[1024];
	      sprintf(msg,"client %d/%d seems to have gone away.\n",l_clients[i],i);
	      syslog(LOG_INFO,msg);
	    }
	    close(l_clients[i]);
	    l_clients[i]=0;
	  }
	} else {
	  /*	  printf("fd %d/%d was not ready for writing\n",l_clients[i],i);*/
	}
      }
    }
  } /*if*/
  if (FD_ISSET(l_fd,&read_fds)) {   /*we have a new TCP client*/
    int l_connfd,new_index=-1;
    socklen_t l_addrlen;
    struct sockaddr * l_cliaddr;

    /*    printf("incoming connection on tcp 4444\n");*/
    l_addrlen=sizeof(l_cliaddr);
    if ( (l_connfd = accept(l_fd,(struct sockaddr *) l_cliaddr, &l_addrlen)) < 0)
      { log_error("accept error"); exit(0); }

    SetTCPOpts(l_connfd);

    for (i=0; i<MAXCLIENTS; i++) {
      if (l_clients[i]==0) {
	if (debug) {
	  printf("New client:  Assigning fd %d to l_clients[%d]\n",l_connfd,i);
	}
	l_clients[i]=l_connfd;
	new_index=i;
	i=MAXCLIENTS;
      }
    }
    if (new_index == -1) {
      char msg[1024];
      printf(msg," Alread talking to %d clients.  New client rejected.\n");
      syslog(LOG_ERR,msg);
      close(l_connfd);
    }
  } /*if*/

  /*check if any tcp clients have disconnected*/
  for(i=0; i<MAXCLIENTS; i++) {
    char junkbuf[4096];
    if ( (l_clients[i]>0) && (FD_ISSET(l_clients[i],&read_fds)) ) {
      /* we should check Recfrom's return value ourselves. */
      c=recvfrom(l_clients[i],junkbuf,sizeof(junkbuf),0,(struct sockaddr *)&l_cliaddr,&l_clilen);
      if (c==0) {
	if (debug) printf("Closing connection %d/%d\n",l_clients[i],i);
	close(l_clients[i]);
	l_clients[i]=0;
      } else if (c>0) { /* if the client sent something, read it and junk it*/
	if (debug) printf("read %d bytes from client %d/%d\n",c,l_clients[i],i);
      } else if (c<0) {
	if (c != EINTR) {
	  if (debug) printf("client %d/%d died somehow.\n",l_clients[i],i);
	  close(l_clients[i]);
	  l_clients[i]=0;
	}
      }
    } /*if*/
  } /*for*/
  
 } /*for ( ; ; )*/
}

 /* Increase our receive buffer for nf_fd: */
int SetUDPOpts(int fd){
  int r,p,n;
  socklen_t       optlen;
  
  n=512*1024;
  optlen = sizeof(p);
  
  if (debug) {
    r=getsockopt(fd,SOL_SOCKET,SO_RCVBUF,&p,&optlen);
    printf("Before setsockopt, SO_RCVBUF is %d\n",p);
  }
  r=setsockopt(fd,SOL_SOCKET,SO_RCVBUF,&n,sizeof(n));
  if (debug) {
    r=getsockopt(fd,SOL_SOCKET,SO_RCVBUF,&p,&optlen);
    printf("After setsockopt, SO_RCVBUF is %d\n",p);
  }
}

int SetTCPOpts(int fd) { 
  int r,p,n;
  socklen_t       optlen;
  
  optlen = sizeof(p);
  
  if (debug) {
    r=getsockopt(fd,SOL_SOCKET,SO_SNDBUF,&p,&optlen);
    printf("Before setsockopt, SO_SNDBUF is %d\n",p);
  }
  n=512*1024;
  r=setsockopt(fd,SOL_SOCKET,SO_SNDBUF,&n,sizeof(n));
  if (r<0) { log_error("setsockopt SO_SNDBUF"); }
  if (debug) {
    r=getsockopt(fd,SOL_SOCKET,SO_SNDBUF,&p,&optlen);
    printf("After setsockopt, SO_SNDBUF is %d\n",p);
  }
  
  if (debug) {
    r=getsockopt(fd,SOL_SOCKET,SO_SNDLOWAT,&p,&optlen);
    if (r<0) { log_error("getsockopt SO_SNDLOWAT"); }
    else {printf("Before setsockopt, SO_SNDLOWAT is %d\n",p); }
  }
  n=2048;
  r=setsockopt(fd,SOL_SOCKET,SO_SNDLOWAT,&n,sizeof(n));
  if (debug) {
    if (r<0) { log_error("setsockopt SO_SNDLOWAT"); }
    r=getsockopt(fd,SOL_SOCKET,SO_SNDLOWAT,&p,&optlen);
    if (r<0) { log_error("getsockopt SO_SNDLOWAT"); }
    else { printf("After setsockopt, SO_SNDLOWAT is %d\n",p); }
  }
  
  if (debug) {
    r=getsockopt(fd,IPPROTO_TCP,TCP_NODELAY,&p,&optlen);
    printf("Before setsockopt, TCP_NODELAY is %d\n",p);
  }
  n=1;
  r=setsockopt(fd,IPPROTO_TCP,TCP_NODELAY,&n,sizeof(n));
  if (debug) {
    if (r<0) { log_error("setsockopt TCP_NODELAY"); }
    r=getsockopt(fd,IPPROTO_TCP,TCP_NODELAY,&p,&optlen);
    printf("After setsockopt, TCP_NODELAY is %d\n",p);
  }
  
}

void log_error(char * errdescr) {

  if (use_syslog) {
    char * errstr;
    char logmsg[1024];
    errstr=(char *)strerror(errno);
    sprintf(logmsg,"%s: %s",errdescr,errstr);
    syslog(LOG_ERR,logmsg);
  } else {
    perror(errdescr);
  }

}
